"""
Describes a raster-like regular grid.
Will replace the Rgrid object at some points
WIP
"""

import numpy as np
import matplotlib.pyplot as plt
from scabbard import io
from scabbard import geometry as geo
import dagger as dag
import scabbard as scb
from scipy.ndimage import gaussian_filter
import random
from copy import deepcopy


class RegularRasterGrid(object):
    """
    Manages a regular grid with helper functions
    """

    def __init__(self, value, geometry, dtype=np.float32):
        """ """

        super().__init__()

        # Check compiance with geometry:
        if isinstance(geometry, geo.RegularGeometry) == False:
            raise AttributeError(
                "a RegularRasterGrid object must be created with a geometry of type RegularGeometry"
            )
            # Checking if the geometry correspond to the shape of the array
            if Z.shape != geometry.shape:
                raise AttributeError(
                    "Matrix not the size indicated in geometry when trying to instanciate RegularRasterGrid"
                )

        # All good then I can instanciate the values
        self.Z = value
        self.geo = geometry

        # Converting type if needed
        if self.Z.dtype != dtype:
            self.Z = self.Z.astype(dtype)

    def duplicate_with_other_data(self, value):
        """
        Returns a copy of this raster with different data
        Useful to create another array with the same geometry and all

        Arguments:
                - value: the 2D numpy array replacing the value

        Returns:
                - A copy of the current raster with the new value field
        Authors:
                - B.G (last modifications 09/2024)
        """

        co = deepcopy(self)
        co.Z = value

        return co

    @property
    def dims(self):
        """
        topotoolbox-friendly dim parameter
        """
        return np.array([self.geo.ny, self.geo.nx], dtype=np.uint64)

    @property
    def rshp(self):
        """
        Arg to feed to np.reshape for 1D->2D
        """
        return np.array([self.geo.ny, self.geo.nx], dtype=np.uint64)

    def grid2ttb(self):
        '''
        Returns a libtopotoolbox-compatible grid
        '''
        import topotoolbox as ttb
        ttbgrid = ttb.GridObject()
        ttbgrid.z = self.Z
        ttbgrid.rows = self.geo.ny
        ttbgrid.columns = self.geo.nx
        ttbgrid.shape = self.Z.shape
        ttbgrid.cellsize = self.geo.dx
        ttbgrid.bounds = self.geo.extent
        ttbgrid.name = "from_scabbard"

        return ttbgrid

    # Common operator overload
    def __add__(self, other):
        if isinstance(other, RegularRasterGrid):
            return RegularRasterGrid(self.Z + other.Z, self.geo)
        else:
            return RegularRasterGrid(self.Z + other, self.geo)

    def __sub__(self, other):
        if isinstance(other, RegularRasterGrid):
            return RegularRasterGrid(self.Z - other.Z, self.geo)
        else:
            return RegularRasterGrid(self.Z - other, self.geo)

    def __mul__(self, other):
        if isinstance(other, RegularRasterGrid):
            return RegularRasterGrid(self.Z * other.Z, self.geo)
        else:
            return RegularRasterGrid(self.Z * other, self.geo)

    def __truediv__(self, other):
        if isinstance(other, RegularRasterGrid):
            return RegularRasterGrid(self.Z / other.Z, self.geo)
        else:
            return RegularRasterGrid(self.Z / other, self.geo)




def raster_from_array(Z, dx=1.0, xmin=0.0, ymin=0.0, dtype=np.float32):
    """
    Helper function to get a RegularRasterGrid from a 2D array

    Arguments:
            - Z: the 2D value
            - dx: the spatial step
            - xmin: left coordinate bound
            - ymin: bottom coordinate bound
            - dtype: numpy data type of Z

    Returns:
            - a RegularRasterGrid object

    Authors:
            - B.G. (last modification 09/2024)
    """
    geometry = scb.geometry.RegularGeometry(Z.shape[1], Z.shape[0], dx, xmin, ymin)
    return RegularRasterGrid(Z, geometry, dtype=dtype)

def raster_from_ttb(ttbgrid):
    '''
    Converts a libtopotoolbox gridObjg to scabbard grid object
    
    Arguments:
        - ttbgrid: the GRidObj from topotoolbox

    Returns:
        - A regularGridObject

    Authors:
        - B.G. (last modification: 11/2024)

    '''

    geometry = scb.geometry.RegularGeometry(ttbgrid.columns, ttbgrid.rows, ttbgrid.cellsize, ttbgrid.bounds[0], ttbgrid.bounds[1])
    return RegularRasterGrid(ttbgrid.z, geometry, dtype=ttbgrid.z.dtype)





'''
Contains routines to create raster grids from scratch.

For example white noise, perlin noise or spectral stuff for initial conditions,
or constant slope raster to test river LEMs, ...

B.G. - created 09/2024
'''

import scabbard as scb
import numpy as np
import matplotlib.pyplot as plt
import taichi as ti
import numba as nb
import time


def slope2D_S(
	nx = 256, 
	ny = 512, 
	dx = 2.,
	z_base = 0,
	slope = 1e-3, 
	):
	'''
		Generates a slopping grid adn initialise an env with given boundaries to it. 
		It comes with a connector and a graph (todo) alredy pre-prepared for boundaries
	'''

	# Zeros Topo
	Z = np.zeros((ny, nx), dtype = np.float32)

	lx = (nx+1) * dx
	ly = (ny+1) * dx

	grid = scb.raster.raster_from_array(Z, dx=dx, xmin=0.0, ymin=0.0, dtype=np.float32)

	XX,YY = grid.geo.XY
	
	Z = (YY * slope)[::-1]
	grid.Z[:,:] = Z[:,:]

	BCs = np.ones((ny,nx),dtype = np.uint8)
	BCs[:,[0,-1]] = 0
	BCs[0,:] = 7
	BCs[-1,:] = 3

	return grid, BCs


#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages
from setuptools.command.install import install
import os

VERSION = '0.0.14'


class PostInstallCommand(install):
	"""Post-installation for installation mode."""
	def run(self):
		# Call superclass's run method
		install.run(self)
		
		# Your post-installation logic here
		os.system("scb-reset-config")
		print("KDSFDSJKFSLDKHF")



with open('README.md') as readme_file:
		readme = readme_file.read()

with open('HISTORY.rst') as history_file:
		history = history_file.read()

requirements = ['Click>=7.0', 'numpy>=2', 'numba>=0.60', 'daggerpy>=0.0.14', 'matplotlib', 'taichi', 'rasterio', 'scipy', 'cmcrameri', 'h5py', 'topotoolbox']

test_requirements = [ ]

setup(
	author="Boris Gailleton",
	author_email='boris.gailleton@univ-rennes.fr',
	python_requires='>=3.10',
	classifiers=[
			'Development Status :: 4 - Beta',
			'License :: OSI Approved :: MIT License',
			'Natural Language :: English',
			'Programming Language :: Python :: 3',
			'Programming Language :: Python :: 3.10',
			'Programming Language :: Python :: 3.11',
			'Programming Language :: Python :: 3.12',
	],
	description="Suite of Hydrodynamic, topographic analysis, Landscape Evolution model and visualisation tools",
	entry_points={

		'console_scripts': [
				'scb-baseplot=scabbard.visu.nice_terrain:cli_nice_terrain' ,
				'scb-crop=scabbard.raster.std_raster_cropper:std_crop_raster' ,
				'scb-graphflood=scabbard.phineas:graphflood_basic' ,
				'scb-reset-config=scabbard.config:defaultConfig' ,
				'scb-visu2D=scabbard.phineas:visu2Dnpy' ,
				'scb-quick-hydro=scabbard.phineas:GPUgraphflood' ,
		],
	},
	
	# scripts = ['scabbard/nice_haguid.py'],

	install_requires=requirements,
	license="MIT license",
	long_description=readme + '\n\n' + history,
    long_description_content_type='text/markdown',  # For Markdown
	include_package_data=True,
	keywords='scabbard',
	name='pyscabbard',
	packages=find_packages(),
	package_data={
		'scabbard': ['data/*.json'],  # Include your config.json file
		# 'scabbard': ['steenbok/*.cu'],  # Include your config.json file
	},
	cmdclass={
		'install': PostInstallCommand,
	},
	test_suite='tests',
	tests_require=test_requirements,
	url='https://github.com/bgailleton/scabbard',
	version=VERSION,
	zip_safe=False,
)



'''
Different graph wrappers for different libraries.


B.G. (last modification: 07/2024)
'''


import numpy as np
import dagger as dag
import scabbard as scb
import scabbard._utils as ut
import topotoolbox as ttb



class SFGraph(object):
	"""
	Simple single flow direction graph container.
	Only contains the minimal structure of the graph:
		- Single receivers and donors indices (row major flat index)
		- Stack (topological ordering)
		- Conversion helpers

	B.G. (last modification: )
	"""

	def __init__(self, Z, BCs = None, D4 = True, dx = 1., backend = 'ttb', fill_LM = False, step_fill = 1e-3):
		'''
		'''
		
		if(len(Z.shape) != 2):
			raise ValueError('Need a 2D array as input for topography to infer the shape')

		# Geometrical infos
		## Overall shape
		self.shape = Z.shape
		self.dim = np.array(self.shape, dtype = np.uint64)
		# Geometrical infos
		self.D4 = D4

		self.dx = dx


		self.backend = backend

		self.gridcpp = dag.GridCPP_f32(self.nx,self.ny, dx, dx,3) 

		# Getting the graph structure ready
		## Steepest receiver flat index
		self.Sreceivers = np.zeros((self.nxy), dtype = (np.uint64 if self.backend == 'ttb' else np.int32))
		## distance to Steepest receiver flat index
		self.Sdx = np.zeros((self.nxy),dtype = np.float32)
		## Number of Steepest donor per node
		self.Ndonors = np.zeros((self.nxy), dtype = np.uint8 if self.backend == 'ttb' else np.int32)
		## Steepest donor per node ([i,j,:Ndonors[i,j]])
		self.donors = np.zeros((self.nxy * (4 if self.D4 else 8) ), dtype = (np.uint64 if self.backend == 'ttb' else np.int32))
		## Topological ordering
		self.Stack = np.zeros(self.nxy, dtype = (np.uint64 if self.backend == 'ttb' else np.int32))

		# Initialising the graph
		self.update(Z, BCs, fill_LM,step_fill)

	

	def update(self, Z, BCs = None, fill_LM = False, step_fill = 1e-3):
		'''
		Updates the graph to a new topography and optional boundary conditions
		'''

		if(Z.shape != self.shape):
			raise AttributeError('topography needs to be same shape as the graph')

		if(BCs is None):
			BCs = ut.normal_BCs_from_shape(self.nx,self.ny)

		if(self.backend == 'dagger'):
			if self.D4:
				dag.compute_SF_stack_D4_full_f32(self.gridcpp, Z, self.Sreceivers.reshape(self.ny,self.nx), self.Ndonors.reshape(self.ny,self.nx), self.donors.reshape(self.ny,self.nx), self.Stack, BCs)
			else:
				raise ValueError('D8 SFGraph not implemented yet')
		elif self.backend == 'ttb':
			self.Ndonors.fill(0)
			#Theer is a bug in my implementation in ttb, I need to sort it but the vanilla fill LM does not fill the flats
			if(fill_LM):
				scb.ttb.graphflood.funcdict['priority_flood_TO'](Z.ravel(), self.Stack, BCs.ravel(), self.dim, not self.D4, step_fill)
				
			ttb.graphflood.funcdict['sfgraph'](Z.ravel(), self.Sreceivers, self.Sdx, self.donors, self.Ndonors, self.Stack, BCs.ravel(), self.dim, self.dx * self.dx, not self.D4, False, step_fill) # False cause I fill lm above


	@property
	def nx(self):
		return self.shape[1]
	@property
	def ny(self):
		return self.shape[0]
	@property
	def nxy(self):
		return self.shape[0]*self.shape[1]
	




import matplotlib.pyplot as plt
import numpy as np
from matplotlib.ticker import FuncFormatter

def m_to_km(ax, ndec = 1):
	'''
	Takes a matplotlib axis and converts the units from m to km
	Cannot really check what the axis units are so effectively just divides them by 1000

	Arguments:
		- ax: the matplotlib axis
		- ndec (int): the number of decimals to round for

	Returns:
		- nothing, edits in place

	Authors:
		- B.G. (last modifications: 08/2024)

	'''

	# Nested function to format the tick labels
	def format_func(value, tick_number):
		return f'{round(value / 1000, ndec)}'

	# Apply the formatting function to x and y axis tick labels
	ax.set_xticklabels([format_func(tick, i) for i, tick in enumerate(ax.get_xticks())])
	ax.set_yticklabels([format_func(tick, i) for i, tick in enumerate(ax.get_yticks())])

	# Eventually changes the axis labels
	if '(m)' in ax.get_xlabel():
		ax.set_xlabel(ax.get_xlabel().replace('(m)','(km)'))
	if '(m)' in ax.get_ylabel():
		ax.set_ylabel(ax.get_ylabel().replace('(m)','(km)'))		



def tickgrid(ax, marker = '+', size = 40, color = 'k', alpha = 0.7):
	'''
	Adds a tick grid to an existing axis
	tickgrid is a grid only plotting markers at the crossings of the grid lines
	Feat. Luca Malatesta.

	Arguments:
		- ax: the matplotlib axis
		- marker: the matplotlib marker type
		- size: the point size (argument s in scatter plots)
		- color: the color (any matplotlib accepted color)
		- alpha: the transparency (0-1)

	Returns:
		- nothing, edits in place

	Authors:
		- B.G. (last modifications: 08/2024)

	'''
	X = []
	Y = []

	for x in list(ax.get_xticks()):
		if(x < ax.get_xlim()[0] or x > ax.get_xlim()[1]):
			continue
		for y in list(ax.get_yticks()):
			if(y < ax.get_ylim()[0] or y > ax.get_ylim()[1]):
				continue
			X.append(x)
			Y.append(y)

	ax.scatter(X, Y, marker = marker, s = size, alpha = alpha, edgecolor = color, facecolor = color, lw = 1)


def convert_log_colorbar_labels_to_scientific(cb):
	'''
	Takes a matplotlib colorbar and converts its label to 10^value

	Arguments:
		- cb: the colorbar object
		- ndec (int): the number of decimals to round for

	Returns:
		- nothing, edits in place

	Authors:
		- B.G. (last modifications: 08/2024)

	'''

	# Define a formatter function to convert log10 values to scientific notation
	def scientific_ticks(x, pos):
		return f"$10^{{{int(x)}}}$"

	# Apply the formatter to the colorbar
	cb.ax.xaxis.set_major_formatter(FuncFormatter(scientific_ticks))
	cb.ax.yaxis.set_major_formatter(FuncFormatter(scientific_ticks))


import rasterio
from rasterio import windows
import matplotlib.pyplot as plt
from matplotlib.widgets import RectangleSelector
import click
import os


@click.command()
@click.argument("dem_path", type=str)
@click.argument("output", type=str)
def std_crop_raster(dem_path, output):

    if os.path.exists(dem_path) == False:
        raise ValueError("File does not Exists")

    # Open the DEM raster
    dataset = rasterio.open(dem_path)

    # Read the raster data (assuming it's a single band raster)
    data = dataset.read(1)

    # Get the extent of the raster [xmin, xmax, ymin, ymax]
    extent = [
        dataset.bounds.left,
        dataset.bounds.right,
        dataset.bounds.bottom,
        dataset.bounds.top,
    ]

    # Plot the raster
    fig, ax = plt.subplots(figsize=(10, 10))
    cax = ax.imshow(data, cmap="terrain", extent=extent)
    ax.set_title("DEM")
    fig.colorbar(cax, ax=ax, label="Elevation (m)")

    # Function to handle the rectangle selection and crop the raster
    def onselect(eclick, erelease):
        # Get the coordinates of the rectangle
        x1, y1 = eclick.xdata, eclick.ydata
        x2, y2 = erelease.xdata, erelease.ydata

        # Create a bounding box with the selected coordinates
        minx, maxx = sorted([x1, x2])
        miny, maxy = sorted([y1, y2])

        print(f"Rectangle selected from ({minx}, {miny}) to ({maxx}, {maxy})")

        # Define the window to read
        window = windows.from_bounds(
            minx, miny, maxx, maxy, transform=dataset.transform
        )

        # Read the data within the window
        data_crop = dataset.read(1, window=window)

        # Get the transform object for the cropped data
        transform_crop = windows.transform(window, dataset.transform)

        # Update metadata for the cropped raster
        out_meta = dataset.meta.copy()
        out_meta.update(
            {
                "driver": "GTiff",
                "height": data_crop.shape[0],
                "width": data_crop.shape[1],
                "transform": transform_crop,
            }
        )

        # Save the cropped raster
        cropped_path = output  # You can change the output path here
        with rasterio.open(cropped_path, "w", **out_meta) as dest:
            dest.write(data_crop, 1)

        print(f"Cropped raster saved as {cropped_path}")

    rect_selector = RectangleSelector(
        ax,
        onselect,
        useblit=True,
        button=[1],  # Left mouse button
        minspanx=5,
        minspany=5,
        spancoords="data",
        interactive=True,
    )

    plt.show()



'''
Routines to calculate drainage area and other similar stuff
'''

import numba as nb
import numpy as np
import scabbard.steenbok as st
from scabbard.flow import SFGraph
import scabbard as scb


@nb.njit()
def _drainage_area_sfg(Stack,Sreceivers, dx = 1., BCs = None):

	A = np.zeros_like(Sreceivers, dtype = np.float32)

	for i in range(Stack.shape[0]):
		node = Stack[Stack.shape[0] - 1 - i]
		rec = Sreceivers[node]

		if(node == rec):
			continue

		A[node] += dx * dx
		A[rec] += A[node]

	return A



def drainage_area(input_data):

	if(isinstance(input_data, SFGraph) == False):
		raise RuntimeError('drainage area WIP, so far requires SFGraph object to calculate')

	return _drainage_area_sfg(input_data.Stack, input_data.Sreceivers, dx = input_data.dx).reshape(input_data.ny,input_data.nx)




@nb.njit()
def _propagate_sfg(Stack,Sreceivers, values, dx = 1., BCs = None):

	A = np.zeros_like(Sreceivers, dtype = np.float32)

	for i in range(Stack.shape[0]):
		node = Stack[Stack.shape[0] - 1 - i]
		rec = Sreceivers[node]

		if(node == rec):
			continue

		A[node] += values[node]
		A[rec] += A[node]

	return A


@nb.njit()
def _propagate_mfd_propS(Z, Stack, values, nx, ny, dx = 1., BCs = None, D4 = True):

	A = np.zeros_like(Z, dtype = np.float32)

	for i in range(Stack.shape[0]):
		node = Stack[Stack.shape[0] - 1 - i]

		
		if(scb.ste.can_out_flat(node,BCs) or scb.ste.is_active_flat(node,BCs) == False):
			continue
		
		A[node] += values[node]

		tot_S = 0.
		nrec = 0
		for k in range(4 if D4 else 8):
			nnode = scb.ste.neighbours_D4_flat(node,k,BCs,nx,ny) if D4 else scb.ste.neighbours_D8_flat(node,k,BCs,nx,ny)
			if(nnode == -1):
				continue

			if(Z[node] <= Z[nnode]):
				continue


			tot_S += (Z[node] - Z[nnode])/(scb.ste.dx_from_k_D4(dx,k) if D4 else scb.ste.dx_from_k_D8(dx,k))
			nrec += 1

		if(tot_S == 0.):
			continue

		for k in range(4 if D4 else 8):
			nnode = scb.ste.neighbours_D4_flat(node,k,BCs,nx,ny) if D4 else scb.ste.neighbours_D8_flat(node,k,BCs,nx,ny)

			if(nnode == -1):
				continue

			if(Z[node] <= Z[nnode]):
				continue

			A[nnode] += A[node] * (Z[node] - Z[nnode])/(scb.ste.dx_from_k_D4(dx,k) if D4 else scb.ste.dx_from_k_D8(dx,k))/tot_S

	return A



def propagate(input_data, input_values, method = 'sfd', BCs = None, D4 = True, fill_LM = False, step_fill = 1e-3):
	'''
	Propagates values with the flow, following a drainage-area-like path

	Arguments:
		- input_data: the topographic data, can be a RegularRasterGrid or for single flow path propagation a SFGraph
		- input_values: a 2D array of input data shape containing the data (e.g. precipitations, sources, ...)
		- method: str, 'sfd' for single flow direction, 'mfd_S' for multiple flow partitionned prop. to the slope, other to come


	'''
		

	if(method.lower() == 'sfd'):
		
		if(isinstance(input_data, scb.raster.RegularRasterGrid)):
			tBCs = scb.flow.get_normal_BCs(input_data) if BCs is None else BCs
			stg = scb.flow.SFGraph(tZ, BCs = tBCs, D4 = True, dx = input_data.geo.dx, backend = 'ttb', fill_LM = fill_LM, step_fill=step_fill)
		elif(isinstance(input_data, SFGraph)):
			stg = input_data
		else:
			raise RuntimeError('drainage area using SFD method requires a SFGraph object or a RegularRasterGrid as input_data')

		return _propagate_sfg(input_data.Stack, input_data.Sreceivers, input_values.ravel(), dx = input_data.dx).reshape(input_data.ny,input_data.nx)

	elif (method.lower() == 'mfd_s'):
		if(isinstance(input_data, scb.raster.RegularRasterGrid)):
			tBCs = scb.flow.get_normal_BCs(input_data) if BCs is None else BCs
		else:
			raise RuntimeError('drainage area using MFD methods requires a RegularRasterGrid as input_data')

		if(fill_LM):
			Stack = np.zeros_like(input_data.Z.ravel(), dtype = np.uint64)
			scb.ttb.graphflood.funcdict['priority_flood_TO']( input_data.Z.ravel(), Stack, BCs.ravel(), input_data.dims, not D4, step_fill)

		else:
			Stack = np.argsort(input_data.Z.ravel()).astype(np.uint64)

		return _propagate_mfd_propS(input_data.Z.ravel(), Stack.ravel(), input_values.ravel(), input_data.geo.nx, input_data.geo.ny, dx = input_data.geo.dx, BCs = tBCs.ravel(), D4 = D4).reshape(input_data.geo.ny,input_data.geo.nx)

	else:
		raise RuntimeError('Supported methods so far: sfd or mfd_S')























































# end of file



import scabbard as scb
import numpy as np
import numba as nb


def compute_flow_distance_from_outlet(
    grid: scb.raster.RegularRasterGrid,
    method: str = "mean",
    BCs=None,
    Stack=None,
    fill_LM=False,
    step_fill=1e-3,
    D8=True,
):

    if method.lower() == "mean":

        if isinstance(grid, scb.raster.RegularRasterGrid):
            tBCs = scb.flow.get_normal_BCs(grid) if BCs is None else BCs
        else:
            raise RuntimeError(
                "compute_flow_distance_from_outlet requires a RegularRasterGrid as grid"
            )

        if Stack is None:
            if fill_LM:
                Stack = np.zeros_like(grid.Z.ravel(), dtype=np.uint64)
                scb.ttb.graphflood.funcdict['priority_flood_TO'](
                    grid.Z.ravel(), Stack, BCs.ravel(), grid.dims, D8, step_fill
                )
            else:
                Stack = np.argsort(grid.Z.ravel()).astype(np.uint64)

        return scb.ste.mean_dist_to_outlet(
            Stack,
            grid.Z.ravel(),
            tBCs.ravel(),
            D8,
            grid.geo.nx,
            grid.geo.ny,
            grid.geo.dx,
        ).reshape(grid.rshp)
    elif method.lower() == "min":

        if isinstance(grid, scb.raster.RegularRasterGrid):
            tBCs = scb.flow.get_normal_BCs(grid) if BCs is None else BCs
        else:
            raise RuntimeError(
                "compute_flow_distance_from_outlet requires a RegularRasterGrid as grid"
            )

        if Stack is None:
            if fill_LM:
                Stack = np.zeros_like(grid.Z.ravel(), dtype=np.uint64)
                scb.ttb.graphflood.funcdict['priority_flood_TO'](
                    grid.Z.ravel(), Stack, BCs.ravel(), grid.dims, D8, step_fill
                )
            else:
                Stack = np.argsort(grid.Z.ravel()).astype(np.uint64)

        return scb.ste.min_dist_to_outlet(
            Stack,
            grid.Z.ravel(),
            tBCs.ravel(),
            D8,
            grid.geo.nx,
            grid.geo.ny,
            grid.geo.dx,
        ).reshape(grid.rshp)
    elif method.lower() == "max":

        if isinstance(grid, scb.raster.RegularRasterGrid):
            tBCs = scb.flow.get_normal_BCs(grid) if BCs is None else BCs
        else:
            raise RuntimeError(
                "compute_flow_distance_from_outlet requires a RegularRasterGrid as grid"
            )

        if Stack is None:
            if fill_LM:
                Stack = np.zeros_like(grid.Z.ravel(), dtype=np.uint64)
                scb.ttb.graphflood.funcdict['priority_flood_TO'](
                    grid.Z.ravel(), Stack, BCs.ravel(), grid.dims, D8, step_fill
                )
            else:
                Stack = np.argsort(grid.Z.ravel()).astype(np.uint64)

        return scb.ste.max_dist_to_outlet(
            Stack,
            grid.Z.ravel(),
            tBCs.ravel(),
            D8,
            grid.geo.nx,
            grid.geo.ny,
            grid.geo.dx,
        ).reshape(grid.rshp)
    else:
        raise RuntimeError("Methods can be: 'mean' or 'min' or 'max'. More will come.")



import scabbard as scb
import numpy as np
import dagger as dag
import matplotlib.pyplot as plt
from cmcrameri import cm
import time

class ModelHelper:

	def __init__(self):
		self.grid = None
		self.gf = None

		self.dx = 1
		self.dy = 1
		self.nx = 40
		self.ny = 200
		self.S0 = 1e-2
		self.P = 15


		self.mannings = 0.033
		self.dt = 1e-3
		self.courant = True
		self.courant_number = 0.01
		self.SFD = False
		self.min_courant_dt = 1e-4
		self.stationary = True


	def init_from_grid(self,grid, S0 = 1e-2):
		

		self.grid = grid
		self.nx = grid.nx
		self.ny = grid.ny
		self.dx = grid.dx
		self.dy = grid.dy
		self.S0 = S0
		self.gf = scb.GraphFlood(self.grid,verbose = False)
		self.gf.flood.set_fixed_slope_at_boundaries(self.S0)
		self.gf.flood.set_partition_method(dag.MFD_PARTITIONNING.PROPOSLOPE)
		self.apply_model_params()
		self.grid.compute_graphcon() if self.grid.graph is None else 0
		self.grid.graph.set_LMR_method(dag.LMR.priority_flood)





	def init_slope_model(self, nx, ny, dx, dy, S0, P):
		self.nx = nx
		self.ny = ny
		self.dx = dx
		self.dy = dy
		self.S0 = S0
		self.P = P
		nodes = np.arange(1,self.nx-1, dtype=np.int32) + self.nx
		hwins = self.P * np.ones_like(nodes)
		self.grid = scb.slope_RGrid(self.nx,self.ny,self.dx,self.dy, slope = self.S0, noise_magnitude=0., EW = "noflow", S = "can_out", N = "noflow")
		self.grid.graph.set_LMR_method(dag.LMR.priority_flood)
		# self.grid.graph.set_LMR_method(dag.LMR.priority_full_MFD)
		self.grid.Z2D[:,[0,-1]] += 10000 # Locking the borders
		self.grid.Z2D[0,:] += 5 # Locking the borders
		# self.grid.Z2D[1:-1,:] -= 0.1
		## graphflood python object
		self.gf = scb.GraphFlood(self.grid,verbose = False)
		self.gf.set_input_discharge(nodes, hwins, asprec=True)
		# IMPORTANT: setting boundary slope to bedrock slope
		self.gf.flood.set_fixed_slope_at_boundaries(self.S0)
		self.gf.flood.set_partition_method(dag.MFD_PARTITIONNING.PROPOSLOPE)
		self.apply_model_params()

	def init_dem_model(self, file_name, sea_level = 0., P = 1e-5, BCs = None):
		

		self.P = P

		self.grid = scb.raster2RGrid(file_name)
		
		self.grid.compute_graphcon()
		
		self.grid.graph.set_LMR_method(dag.LMR.priority_flood)

		self.nx = self.grid.nx
		self.ny = self.grid.ny
		self.dx = self.grid.dx
		self.dy = self.grid.dy
		self.S0 = 0.

		if(sea_level != 0. and  BCs is not None):
			print("warning: Ignoring sea level as you gave custom boundary condistions. The latter will prevail.")

		# if(BCs is None ):
		# 	BCs = np.zeros((self.ny,self.nx), dtype = np.uint8) + 1
		# 	BCs[[0,-1],:] = 3
		# 	BCs[:, [0,-1]] = 3
		# 	BCs[BCs <= 0] = 0
			
		
		if(BCs is None):
			dag.set_BC_to_remove_seas(self.grid.con, self.grid._Z, sea_level)
		else:
			self.grid.con.set_custom_boundaries(BCs.ravel())



		## graphflood python object
		self.gf = scb.GraphFlood(self.grid,verbose = False)
		# IMPORTANT: setting boundary slope to bedrock slope
		self.gf.flood.set_fixed_slope_at_boundaries(self.S0)
		self.gf.flood.set_partition_method(dag.MFD_PARTITIONNING.PROPOSLOPE)
		self.gf.set_precipitations(P)
		self.apply_model_params()

	def apply_model_params(self):
		self.gf.flood.set_mannings(self.mannings)	
		self.gf.flood.set_dt_hydro(self.dt)
		if(self.courant):
			self.gf.flood.enable_courant_dt_hydro()
			self.gf.flood.set_courant_number(self.courant_number)
			self.gf.flood.set_min_courant_dt_hydro(self.min_courant_dt)
		if(self.SFD):
			self.gf.flood.enable_SFD()
		else:
			self.gf.flood.enable_MFD()

		if(self.stationary):
			self.gf.flood.enable_hydrostationary()
		else:
			self.gf.flood.disable_hydrostationary()

	@property
	def mannings(self):
		return self._mannings

	@mannings.setter
	def mannings(self,val):
		self._mannings = val
		if(self.gf is not None):
			self.gf.flood.set_mannings(self.mannings)

	@property
	def dt(self):
		return self._dt
	@dt.setter
	def dt(self,val):
		self._dt = val
		if(self.gf is not None):
			self.gf.flood.set_dt_hydro(self.dt)

	@property
	def courant(self):
		return self._courant

	@courant.setter
	def courant(self,val):
		self._courant = val
		if(self.gf is not None and val == True):
			# print("enabling courant")
			self.gf.flood.enable_courant_dt_hydro()
			self.gf.flood.set_courant_number(self._courant_number)
			self.gf.flood.set_min_courant_dt_hydro(self.min_courant_dt)
		elif (self.gf is not None):
			self.gf.flood.disable_courant_dt_hydro()

	@property
	def courant_number(self):
		return self._courant_number
		
	@courant_number.setter
	def courant_number(self,val):
		self._courant_number = val
		if(self.gf is not None):
			self.gf.flood.set_courant_number(self._courant_number)

	@property
	def min_courant_dt(self):
		return self._min_courant_dt
		
	@min_courant_dt.setter
	def min_courant_dt(self,val):
		self._min_courant_dt = val
		if(self.gf is not None):
			# print("DEBUG:setting min courant dt to ", self._min_courant_dt)
			self.gf.flood.set_min_courant_dt_hydro(self._min_courant_dt)

	@property
	def SFD(self):
		return self._SFD
	@SFD.setter
	def SFD(self,val):
		self._SFD = val
		if(self.gf is not None):
			if(self.SFD):
				self.gf.flood.enable_SFD()
			else:
				self.gf.flood.enable_MFD()

	@property
	def P(self):
		return self._P
	@P.setter
	def P(self,val):
		self._P = val
		if(self.gf is not None):
			self.gf.set_precipitations(val)

	@property
	def hw(self):

		ret = self.gf.flood.get_hw()
		if(ret.shape[0] > 0):
			ret = ret.reshape(self.rshp)
			return ret
		return np.zeros(self.rshp)

			

	@property
	def stationary(self):
		return self._stationary

	@stationary.setter
	def stationary(self,val):
		
		self._stationary = val

		if(self.gf is not None):
			if(self.stationary):
				self.gf.flood.enable_hydrostationary()
			else:
				self.gf.flood.disable_hydrostationary()

	@property
	def rshp(self):
		return (self.ny,self.nx)


	def run(self):
		self.gf.flood.run()




class PlotHelper(object):
	"""docstring for PlotHelper"""
	def __init__(self, mod, jupyter = False):
		super(PlotHelper, self).__init__()
		self.mod = mod
		self.figs = []
		self.figaxel = {}
		self.hw_plot = False \

		self.jupyter = jupyter	



	def init_hw_plot(self, alpha_hw = 0.75, cutoff = 0.1, cmap = "Blues", vmin = 0.1, vmax = 1.5, kwargs_subplots = {}, use_extent = True):

		self.hw_plot = True

		fig,ax = plt.subplots(*kwargs_subplots)
		textent = self.mod.grid.extent() if use_extent else None

		imhs = ax.imshow(self.mod.grid.hillshade, extent = textent, vmin =0, vmax = 1, cmap = cm.grayC)
		imhw = ax.imshow(self.mod.hw, extent = textent, vmin =vmin, vmax = vmax, cmap = "Blues", alpha = alpha_hw)
		self.figs.append(fig)
		self.figaxel["fig_hw"] = fig
		self.figaxel["ax_hw"] = ax
		self.figaxel["hs_hw"] = imhs
		self.figaxel["hw_hw"] = imhw
		if(self.jupyter == False):
			fig.show()

	def update(self):

		if(self.hw_plot):
			self.figaxel["hs_hw"].set_data(self.mod.grid.hillshade)
			self.figaxel["hw_hw"].set_data(self.mod.hw)

		if(self.jupyter == False):
			for fig in self.figs:
				fig.canvas.draw_idle()
				fig.canvas.start_event_loop(0.001)
		else:
			for fig in self.figs:
				fig.canvas.draw()
				time.sleep(0.001)

			


